import pydantic
import pytest
from tinytable.table import Table
from pydantable import table
from pydantable.errors import ValidationErrors


class PeopleTable(table.BaseTableModel):
    id: int
    name: str
    age: int


class Ages(pydantic.BaseModel):
    id: int
    age: float


def test_read_csv() -> None:
    tbl: Table = PeopleTable.read_csv('tests/data/people.csv')
    assert tbl.data == {'id': [1, 2, 3], 'name': ['Odos', 'Kayla', 'Dexter'], 'age': [38, 31, 2]}


def test_validation() -> None:
    tbl = Table({'id': ['1', '2'], 'age': ['Six', '99']})
    with pytest.raises(ValidationErrors):
        table.validate_table(tbl, Ages)



from __future__ import annotations
import typing as _t
import pydantic as _pydantic
import tinytable as _tt
import tinytable.csv
import tinytim.rows

from pydantable import errors


def validate_row(
    row: dict,
    model: _t.Type[_pydantic.BaseModel]
) -> dict:
    return dict(model.model_validate(dict(row)))


def validate_table(
    table: _tt.Table,
    model: _t.Type[_pydantic.BaseModel]
) -> _tt.Table:
    out_table: _tt.Table = table.copy()
    validation_errors: list[dict] = []
    errored = False
    for i, row in table.iterrows():
        try:
            validated_row: dict = validate_row(dict(row), model)
            if not errored:
                out_table[i] = validated_row
        except _pydantic.ValidationError as e:
            validation_errors.extend(e.errors())
            errored = True
    if validation_errors:
        grouped_errors: list[dict] = errors.group_errors(validation_errors)
        raise errors.ValidationErrors(grouped_errors)
    return out_table
    

class BaseTableModel(_pydantic.BaseModel):
    # TODO: Add __init__ for reading dict
    
    @classmethod
    def read_csv(cls, path: str) -> _tt.Table:
        tbl: _tt.Table = _tt.read_csv(path)
        return validate_table(tbl, cls)
    
    @classmethod
    def read_dict(cls, d: dict[str, _t.Sequence]) -> _tt.Table:
        tbl = _tt.Table(d)
        return validate_table(tbl, cls)
    
    @classmethod
    def read_excel(cls, path: str, sheet_name: str | None = None) -> _tt.Table:
        tbl: _tt.Table = _tt.read_excel(path, sheet_name)
        return validate_table(tbl, cls)
    
    @classmethod
    def read_sqlite(cls, path: str, table_name: str) -> _tt.Table:
        tbl: _tt.Table = _tt.read_sqlite(path, table_name)
        return validate_table(tbl, cls)
        
    @classmethod
    def read_csv_chunks(cls, path: str, chunksize: int) -> _t.Generator[_tt.Table, None, None]:
        for d_chunk in tinytable.csv.chunk_csv_file(path, chunksize):
            tbl = _tt.Table(d_chunk)
            yield validate_table(tbl, cls)


from collections import defaultdict


class ValidationErrors(Exception):
    ...
    

def combine_errors(errors) -> dict:
    if len(errors) == 0: return {}
    inputs = [error['input'] for error in errors]
    first = errors[0].copy()
    del first['input']
    first['inputs'] = inputs
    return first

def group_errors(errors) -> list[dict]:
    e_data = defaultdict(list)
    for error in errors:
        col = error['loc'][0]
        tp = error['type']
        value = error.copy()
        e_data[(col, tp)].append(value)
        
    for key in e_data.keys():
        e_data[key] = combine_errors(e_data[key])

    return list(e_data.values())